@@ -270,8 +270,6 @@ class AutoFedAvgModel(AutoModel):
                         scheduler_impl: Type['AutoFedAvgScheduler'],
                         is_initiator: bool = False,
                         recover: bool = False,
-                        min_clients: int = 0,
-                        max_clients: int = 0,
                         max_rounds: int = 0,
                         merge_epochs: int = 1,
                         calculation_timeout: int = 300,
@@ -290,18 +288,10 @@ class AutoFedAvgModel(AutoModel):
 
         task_contractor = AutoTaskContractor(task_id=task_id)
         self.clients = task_contractor.query_nodes()
-        involve_aggregator = bool(self.validation_loader
-                                  and len(self.validation_loader) > 0)
-        if involve_aggregator:
-            min_clients = min_clients or len(self.clients)
-            max_clients = max_clients or len(self.clients)
-        else:
-            min_clients = min_clients or len(self.clients) - 1
-            max_clients = max_clients or len(self.clients) - 1
+        involve_aggregator = bool(self.training_loader
+                                  and len(self.training_loader) > 0)
 
         self.scheduler = scheduler_impl(auto_proxy=self,
-                                        min_clients=min_clients,
-                                        max_clients=max_clients,
                                         max_rounds=max_rounds,
                                         merge_epochs=merge_epochs,
                                         calculation_timeout=calculation_timeout,
@@ -322,8 +312,6 @@ class AutoFedAvgScheduler(FedAvgScheduler):
 
     def __init__(self,
                  auto_proxy: AutoFedAvgModel,
-                 min_clients: int,
-                 max_clients: int,
                  max_rounds: int = 0,
                  merge_epochs: int = 1,
                  calculation_timeout: int = 300,
@@ -331,9 +319,7 @@ class AutoFedAvgScheduler(FedAvgScheduler):
                  log_rounds: int = 0,
                  involve_aggregator: bool = False,
                  **kwargs):
-        super().__init__(min_clients=min_clients,
-                         max_clients=max_clients,
-                         max_rounds=max_rounds,
+        super().__init__(max_rounds=max_rounds,
                          merge_epochs=merge_epochs,
                          calculation_timeout=calculation_timeout,
                          schedule_timeout=schedule_timeout,
@@ -399,16 +385,8 @@ class AutoFedAvgScheduler(FedAvgScheduler):
         """
         self.push_log('Uploading task achievement and closing task ...')
 
-        # TODO remove metrics_files later
-        metrics_files = []
-        for _name, _metrics in self._metrics_bucket.items():
-            _file = f'{os.path.join(self._result_dir, _name)}.csv'
-            _metrics.to_csv(_file)
-            metrics_files.append(_file)
         report_file = os.path.join(self._result_dir, "report.zip")
         with ZipFile(report_file, 'w') as report_zip:
-            for _file in metrics_files:
-                report_zip.write(_file, os.path.basename(_file))
             for path, _, filenames in os.walk(self._log_dir):
                 rel_dir = os.path.relpath(path=path, start=self._result_dir)
                 rel_dir = rel_dir.lstrip('.')  # ./file => file
