@@ -178,7 +178,7 @@ class SimpleFedAvgScheduler(Scheduler, metaclass=ABCMeta):
             self._launch_process()
         except Exception:
             # 将错误信息推送到 Playground 前端界面，有助于了解错误原因并修正
-            err_stack = ''.join(traceback.format_exception(*sys.exc_info()))
+            err_stack = '\n'.join(traceback.format_exception(*sys.exc_info()))
             self.push_log(err_stack)
 
     def _recover_progress(self):
@@ -443,7 +443,6 @@ class SimpleFedAvgScheduler(Scheduler, metaclass=ABCMeta):
 
         os.makedirs(self._result_dir, exist_ok=True)
 
-        # TODO remove metrics_files later
         report_file = os.path.join(self._result_dir, "report.zip")
         with ZipFile(report_file, 'w') as report_zip:
             for path, _, filenames in os.walk(self._log_dir):
