@@ -3,14 +3,11 @@
 import io
 import json
 import os
-import random
 import shutil
 import sys
 import traceback
 from abc import ABCMeta, abstractmethod
-from collections import deque
-from dataclasses import dataclass
-from typing import Deque, Dict, List, Tuple
+from typing import Any, Dict, List, Tuple, final
 from zipfile import ZipFile
 
 import torch
@@ -21,15 +18,11 @@ from torch.utils.tensorboard import SummaryWriter
 
 from .. import get_result_dir, get_runtime_dir, logger
 from ..data_channel import SharedFileDataChannel
-from ..perf_bench import SimplePerfBench, SimplePerfBenchResult
 from ..scheduler import ConfigError, Scheduler, TaskComplete, TaskFailed
-from .contractor import (BenchmarkDoneEvent, CheckinEvent,
-                         CheckinResponseEvent, CloseAggregatorElectionEvent,
-                         CloseRoundEvent, DistributeParametersEvent,
-                         FedAvgContractor, FinishTaskEvent,
-                         ReadyForAggregationEvent, ResetRoundEvent,
-                         StartAggregatorElectionEvent, StartRoundEvent,
-                         SynchronizeAggregatorContextEvent, SyncStateEvent,
+from .contractor import (CheckinEvent, CheckinResponseEvent, CloseRoundEvent,
+                         DistributeParametersEvent, FedAvgContractor,
+                         FinishTaskEvent, ReadyForAggregationEvent,
+                         ResetRoundEvent, StartRoundEvent, SyncStateEvent,
                          SyncStateResponseEvent, UploadTrainingResultsEvent)
 
 __all__ = [
@@ -50,42 +43,6 @@ class ResetRound(Exception):
     ...
 
 
-LEN_BYTES = 4
-INT_BYTES_ORDER = 'big'
-
-
-@dataclass
-class AggregatorSyncData:
-
-    state_dict: bytes
-    participants: List[str]
-
-    def to_bytes(self) -> bytes:
-        assert self.state_dict, f'must specify state_dict: {self.state_dict}'
-        assert self.participants, f'must specify participants: {self.participants}'
-
-        data_stream = b''
-        state_dict_len = len(self.state_dict)
-        data_stream += state_dict_len.to_bytes(LEN_BYTES, INT_BYTES_ORDER)
-        data_stream += self.state_dict
-        data_stream += json.dumps(self.participants).encode()
-        return data_stream
-
-    @classmethod
-    def from_bytes(self, data: bytes) -> 'AggregatorSyncData':
-        assert data and isinstance(data, bytes), f'invalid data: {data}'
-
-        archor = 0
-        state_dict_len = int.from_bytes(data[archor:(archor + LEN_BYTES)],
-                                        INT_BYTES_ORDER)
-        archor += LEN_BYTES
-        state_dict = data[archor:(archor + state_dict_len)]
-        archor += state_dict_len
-        participants_bytes = data[archor:]
-        participants = json.loads(participants_bytes.decode())
-        return AggregatorSyncData(state_dict=state_dict, participants=participants)
-
-
 class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
     """Implementation of FedAvg."""
 
@@ -93,9 +50,6 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
     _GETHORING = 'gethering'
     _READY = 'ready'
     _SYNCHRONIZING = 'synchronizing'
-    _ELECTING = 'electing'
-    _BENCHMARKING = 'benchmarking'
-    _SELECTING = 'selecting'
     _IN_A_ROUND = 'in_a_round'
     _UPDATING = 'updating'
     _CALCULATING = 'calculating'
@@ -106,26 +60,15 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
     _FINISHING = 'finishing'
 
     def __init__(self,
-                 min_clients: int,
-                 max_clients: int,
-                 name: str = None,
                  max_rounds: int = 0,
                  merge_epochs: int = 1,
                  calculation_timeout: int = 300,
-                 perf_bench_timeout: int = 30,
                  schedule_timeout: int = 30,
                  log_rounds: int = 0,
-                 is_centralized: bool = True,
                  involve_aggregator: bool = False):
         """Init.
 
         Args:
-            min_clients:
-                Minimal number of calculators for each round.
-            max_clients:
-                Maximal number of calculators for each round.
-            name:
-                Default to the task ID.
             max_rounds:
                 Maximal number of training rounds.
             merge_epochs:
@@ -133,18 +76,12 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
             calculation_timeout:
                 Seconds to timeout for calculation in a round. Takeing off timeout
                 by setting its value to 0.
-            perf_bench_timeout:
-                Seconds to timeout for performing performance benchmark. Takeing off timeout
-                by setting its value to 0.
             schedule_timeout:
                 Seconds to timeout for process scheduling. Takeing off timeout
                 by setting its value to 0.
             log_rounds:
                 The number of rounds to run testing and log the result. Skip it
                 by setting its value to 0.
-            is_centralized:
-                If specify centralized, the aggregator will always be the initiator of the
-                task, otherwize a new aggregator is elected for each round during training.
             involve_aggregator:
                 If set true, the aggregator should have its local data and conduct
                 training other than merely schedule and aggregate.
@@ -152,32 +89,21 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         super().__init__()
         self._switch_status(self._INIT)
 
-        self.name = name
-        self.min_clients = min_clients
-        self.max_clients = max_clients
         self.max_rounds = max_rounds
         self.merge_epochs = merge_epochs
         self.calculation_timeout = calculation_timeout
-        self.perf_bench_timeout = perf_bench_timeout
         self.schedule_timeout = schedule_timeout
         self.log_rounds = log_rounds
-        self.is_centralized = is_centralized
         self.involve_aggregator = involve_aggregator
 
         self._validate_config()
         self.current_round = 1
 
-        self._participants: List[str] = []
-        self._calculators: List[str] = []
-        self._perf_benchmarks: List[SimplePerfBenchResult] = []
-        self._latest_aggregators: Deque[str] = deque()
-        self._is_gathering_complete = False
-
-        self._example_inputs = None
+        self.participants: List[str] = []
+        self.gethered: List[str] = []
+        self.is_gathering_complete = False
 
     def _validate_config(self):
-        if self.min_clients > self.max_clients:
-            raise ConfigError('min_clients must not exceed max_clients')
         if self.merge_epochs <= 0:
             raise ConfigError('merge_epochs must be a positive integer')
 
@@ -185,6 +111,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
     def build_model(self) -> Module:
         """Return a model object which will be used for training."""
 
+    @final
     @property
     def model(self) -> Module:
         """Get the model object which is used for training."""
@@ -201,6 +128,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
                 The model object which is used for training.
         """
 
+    @final
     @property
     def optimizer(self) -> Optimizer:
         """Get the optimizer object which is used for training."""
@@ -218,6 +146,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
             training dataloader
         """
 
+    @final
     @property
     def train_loader(self) -> DataLoader:
         """Get the training dataloader object."""
@@ -235,6 +164,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         """
         raise NotImplementedError()
 
+    @final
     @property
     def validation_loader(self) -> DataLoader:
         """Get the validation dataloader object if needed."""
@@ -256,6 +186,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
             testing dataloader
         """
 
+    @final
     @property
     def test_loader(self) -> DataLoader:
         """Get the testing dataloader object."""
@@ -263,22 +194,24 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
             self._test_loader = self.build_test_dataloader()
         return self._test_loader
 
-    @abstractmethod
     def state_dict(self) -> Dict[str, torch.Tensor]:
         """Get the params that need to train and update.
 
         Only the params returned by this function will be updated and saved during aggregation.
+        Return self.model.state_dict() by default.
 
         Return:
             List[torch.Tensor], The list of model params.
         """
+        return self.model.state_dict()
 
-    @abstractmethod
     def load_state_dict(self, state_dict: Dict[str, torch.Tensor]):
         """Load the params that trained and updated.
 
         Only the params returned by state_dict() should be loaded by this function.
+        Execute self.model.load_state_dict(state_dict) by default.
         """
+        self.model.load_state_dict(state_dict)
 
     def validate_context(self):
         """Validate if the local running context is ready.
@@ -295,11 +228,11 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
             raise ConfigError('Support torch.optim.Optimizer only')
 
     @abstractmethod
-    def train_an_epoch(self):
+    def train_an_epoch(self) -> Any:
         """Define the training steps in an epoch."""
 
     @abstractmethod
-    def test(self):
+    def run_test(self) -> Any:
         """Define the testing steps.
 
         If you do not want to do testing after training, simply make it pass.
@@ -319,7 +252,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
                 self._clean_progress()
             self._launch_process()
         except Exception:
-            err_stack = ''.join(traceback.format_exception(*sys.exc_info()))
+            err_stack = '\n'.join(traceback.format_exception(*sys.exc_info()))
             self.push_log(err_stack)
 
     def _setup_context(self, id: str, task_id: str, is_initiator: bool = False):
@@ -328,8 +261,6 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
 
         self.id = id
         self.task_id = task_id
-        if not self.name:
-            self.name = self.task_id
         self._runtime_dir = get_runtime_dir(self.task_id)
         self._context_file = os.path.join(self._runtime_dir, ".context.json")
         self._checkpoint_dir = os.path.join(self._runtime_dir, 'checkpoint')
@@ -340,29 +271,26 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
 
         self.is_initiator = is_initiator
         self.is_aggregator = self.is_initiator
-        if self.involve_aggregator:
-            self._participants.append(self.id)
 
         self.contractor = FedAvgContractor(task_id=task_id)
         self.data_channel = SharedFileDataChannel(self.contractor)
         self.model
         self.optimizer
 
+        self.participants = self.contractor.query_nodes()
+        if len(self.participants) < 2:
+            raise TaskFailed('Failed to get participants list.')
+
         self.push_log(message='Begin to validate local context.')
         self.validate_context()
 
-        def save_example_inputs(module, input, output):
-            if self._example_inputs is None:
-                self._example_inputs = input
-        self.model.register_forward_hook(save_example_inputs)
-
     def _recover_progress(self):
         """Try to recover and continue from last running."""
         if not os.path.isfile(self._context_file):
             raise TaskFailed('Failed to recover progress: missing cached context.')
 
         with open(self._context_file, 'r') as f:
-            context_info = json.load(f)
+            context_info: dict = json.load(f)
         round = context_info.get('round')
         ckpt_file = context_info.get('ckpt_file')
         assert round and isinstance(round, int) and round > 0, f'Invalid round: {round} .'
@@ -403,11 +331,6 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
                     self._switch_status(self._SYNCHRONIZING)
                     self._sync_state()
 
-                    if not self.is_centralized:
-                        # elect an aggregator for this round
-                        self._switch_status(self._ELECTING)
-                        self._elect_aggregator()
-
                     self._switch_status(self._IN_A_ROUND)
                     self._run_a_round()
                     self._switch_status(self._READY)
@@ -439,7 +362,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         if self.is_initiator:
             self.push_log('Waiting for participants taking part in ...')
             self._wait_for_gathering()
-            self._is_gathering_complete = True
+            self.is_gathering_complete = True
         else:
             is_checked_in = False
             # the aggregator may be in special state so can not response
@@ -457,42 +380,13 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         """
         self.push_log('Synchronizing round state ...')
         if self.is_aggregator:
-            self._calculators.clear()
-            if self.involve_aggregator:
-                self._calculators.append(self.id)
             self.push_log(f'Initiate state synchronization of round {self.current_round}.')
             self.contractor.sync_state(round=self.current_round, aggregator=self.id)
             self._wait_for_sync_response()
-            if len(self._calculators) < self.min_clients:
-                self.push_log('Task failed because of too few participants.')
-                raise AggregationError(f'too few participants: {len(self._calculators)}')
         else:
             self._wait_for_sync_state()
         self.push_log(f'Successfully synchronized state in round {self.current_round}')
 
-    def _elect_aggregator(self):
-        """Elect a new aggregator for a round.
-
-        As an aggregator, initiates election and selects an appropriate one as
-        the new aggregator depending on benchmark result, then transfers context
-        data to the new aggregator.
-        As a participant, performes benchmark and reports the result. If is selected
-        as the new aggregator, receives context data and takes over management.
-        """
-        self.push_log('Electing a new aggregator.')
-        if self.is_aggregator:
-            self.contractor.start_aggregator_election(round=self.current_round)
-            new_aggregator = self._wait_for_benchmark()
-            self.contractor.close_aggregator_election(round=self.current_round)
-            # it's possible the all benchmark failed
-            self.is_aggregator = new_aggregator == self.id
-        else:
-            self.push_log('Waiting for aggregator election begin ...')
-            self._wait_for_electing()
-            self.push_log('Waiting for aggregator election result ...')
-            self._wait_for_election_result()
-        self.push_log('Aggregator election complete.')
-
     def _run_a_round(self):
         """Perform a round of FedAvg calculation.
 
@@ -506,7 +400,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
             try:
                 self._run_as_aggregator()
             except AggregationError as err:
-                err_stack = ''.join(traceback.format_exception(*sys.exc_info()))
+                err_stack = '\n'.join(traceback.format_exception(*sys.exc_info()))
                 self.push_log(err_stack)
                 self.contractor.reset_round()
                 raise ResetRound(err)
@@ -521,7 +415,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         As a participant, do nothing.
         """
         self.push_log(f'Closing task {self.task_id} ...')
-        if self.is_aggregator:
+        if is_succ and self.is_aggregator:
             self._switch_status(self._FINISHING)
             self.contractor.finish_task()
             report_file_path, model_file_path = self._prepare_task_output()
@@ -529,6 +423,9 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
                                                    report_file=report_file_path,
                                                    model_file=model_file_path)
             self.contractor.notify_task_completion(result=True)
+        elif self.is_aggregator:
+            self.contractor.finish_task()
+            self.contractor.notify_task_completion(result=False)
         self.push_log(f'Task {self.task_id} closed. Byebye!')
 
     def _wait_for_gathering(self):
@@ -537,38 +434,33 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         for _event in self.contractor.contract_events():
             if isinstance(_event, CheckinEvent):
                 self._handle_check_in(_event)
-                if len(self._participants) >= self.min_clients:
-                    break
-        # it takes seconds to send a response and to consume it in the remote side,
-        # meanwhile, more check in events may arrive soonly. So keep looking for a while.
-        countdown = 10
-        for _event in self.contractor.contract_events(timeout=countdown):
-            if isinstance(_event, CheckinEvent):
-                self._handle_check_in(_event)
+                if len(self.gethered) == len(self.participants) - 1:
+                    return
 
     def _handle_check_in(self, _event: CheckinEvent):
-        if _event.peer_id not in self._participants:
-            self._participants.append(_event.peer_id)
+        if _event.peer_id not in self.gethered:
+            self.gethered.append(_event.peer_id)
             self.push_log(f'Welcome a new participant ID: {_event.peer_id}.')
-            self.push_log(f'There are {len(self._participants)} participants now.')
+            self.push_log(f'There are {len(self.gethered) + 1} participants now.')
         self.contractor.respond_check_in(round=self.current_round,
                                          aggregator=self.id,
                                          nonce=_event.nonce,
                                          requester_id=_event.peer_id)
-        if self._is_gathering_complete:
+        if self.is_gathering_complete:
             self.contractor.sync_state(round=self.current_round, aggregator=self.id)
 
     def _wait_for_sync_response(self):
         """Wait for participants' synchronizing state response."""
         self.push_log('Waiting for synchronization responses ...')
+        synced = set()
         for _event in self.contractor.contract_events(timeout=0):
             if isinstance(_event, SyncStateResponseEvent):
                 if _event.round != self.current_round:
                     continue
-                self._calculators.append(_event.peer_id)
+                synced.add(_event.peer_id)
                 self.push_log(f'Successfully synchronized state with ID: {_event.peer_id}.')
-                self.push_log(f'Successfully synchronized with {len(self._calculators)} participants.')
-                if len(self._calculators) == len(self._participants):
+                self.push_log(f'Successfully synchronized with {len(synced)} participants.')
+                if len(synced) == len(self.gethered):
                     return
 
             elif isinstance(_event, CheckinEvent):
@@ -613,94 +505,6 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
             elif isinstance(_event, FinishTaskEvent):
                 raise TaskComplete()
 
-    def _wait_for_benchmark(self) -> str:
-        """Wait for someone complete benchmark task.
-
-        Return:
-            the ID of the newly selected aggregator
-        """
-        self.push_log('Waiting for benchmark results ...')
-        new_aggregator = self.id
-        self._calculators.clear()
-        if self.involve_aggregator:
-            self._calculators.append(self.id)
-        for _event in self.contractor.contract_events(timeout=SimplePerfBench.TIMEOUT):
-            if isinstance(_event, BenchmarkDoneEvent):
-                if _event.round != self.current_round:
-                    continue
-
-                if _event.peer_id not in self._calculators:
-                    self._calculators.append(_event.peer_id)
-                    self.push_log(f'Received benchmark result from ID: {_event.peer_id}.')
-
-                # the first one arrived
-                if new_aggregator == self.id:
-                    new_aggregator = _event.peer_id
-                    self.push_log(f'ID: {new_aggregator} is elected as the new aggregator.')
-                    self.push_log('Begin to synchronize context with the new aggregatro.')
-
-                    buffer = io.BytesIO()
-                    torch.save(self.state_dict(), buffer)
-                    # if self.involve_aggregator is set false, add self into participants list
-                    if not self.involve_aggregator:
-                        self._participants.append(self.id)
-                    sync_data = AggregatorSyncData(state_dict=buffer.getvalue(),
-                                                   participants=self._participants)
-                    self.data_channel.send_stream(source=self.id,
-                                                  target=[_event.peer_id],
-                                                  data_stream=sync_data.to_bytes())
-                    self.push_log('Successfully synchronized context with the new aggregatro.')
-
-                if len(self._calculators) == len(self._participants):  # all finished
-                    break
-
-            elif isinstance(_event, CheckinEvent):
-                self._handle_check_in(_event)
-
-        # it's possible that there is no one complete benchmarking
-        return new_aggregator
-
-    def _wait_for_electing(self):
-        """Try for being an aggregator."""
-        logger.debug('_wait_for_electing ...')
-        for _event in self.contractor.contract_events():
-            if isinstance(_event, StartAggregatorElectionEvent):
-                if _event.round != self.current_round:
-                    continue
-                SimplePerfBench().run()
-                self.contractor.report_benchmark_done(round=self.current_round, peer_id=self.id)
-                return
-            elif isinstance(_event, FinishTaskEvent):
-                raise TaskComplete()
-
-    def _wait_for_election_result(self):
-        logger.debug('_wait_for_election_result ...')
-        for _event in self.contractor.contract_events():
-            if isinstance(_event, SynchronizeAggregatorContextEvent):
-                # is selected as the next aggregator and synchronise aggregator data
-                self.push_log('Current node is selected as the new aggregator.')
-                self.push_log('Begin to synchronize context with the original aggregator.')
-                data_stream = self.data_channel.receive_stream(apply_event=_event,
-                                                               receiver=self.id)
-                sync_data = AggregatorSyncData.from_bytes(data=data_stream)
-                buffer = io.BytesIO(sync_data.state_dict)
-                _new_state_dict = torch.load(buffer)
-                self.load_state_dict(state_dict=_new_state_dict)
-                self._participants = sync_data.participants
-                if self.id in self._participants and not self.involve_aggregator:
-                    self._participants.remove(self.id)
-                self.is_aggregator = True
-                self.push_log('Successfully synchronized context with the original aggregator.')
-
-            elif isinstance(_event, CloseAggregatorElectionEvent):
-                if _event.round != self.current_round:
-                    continue
-                else:
-                    return
-
-            elif isinstance(_event, FinishTaskEvent):
-                raise TaskComplete()
-
     def _run_as_aggregator(self):
         self._start_round()
         self._distribute_model()
@@ -719,12 +523,8 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
     def _start_round(self):
         """Prepare and start calculation of a round."""
         self.push_log(f'Begin the training of round {self.current_round}.')
-        self._switch_status(self._SELECTING)
-        self.push_log('Select calculators of current round.')
-        self._select_calculators()
-        self.push_log(f'Calculators of round {self.current_round} are: {self._calculators}.')
         self.contractor.start_round(round=self.current_round,
-                                    calculators=self._calculators,
+                                    calculators=self.gethered,
                                     aggregator=self.id)
         self.push_log(f'Calculation of round {self.current_round} is started.')
 
@@ -732,19 +532,18 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         buffer = io.BytesIO()
         torch.save(self.state_dict(), buffer)
         self.push_log('Distributing parameters ...')
-        results = {_target: False for _target in self._calculators}
-        targets = [_target for _target in results.keys() if _target != self.id]
+        results = {_target: False for _target in self.gethered}
         accept_list = self.data_channel.send_stream(source=self.id,
-                                                    target=targets,
+                                                    target=list(results.keys()),
                                                     data_stream=buffer.getvalue())
         self.push_log(f'Successfully distributed parameters to: {accept_list}')
-        if len(targets) != len(accept_list):
-            reject_list = [_target for _target in targets if _target not in accept_list]
+        if len(results.keys()) != len(accept_list):
+            reject_list = [_target for _target in results.keys()
+                           if _target not in accept_list]
             self.push_log(f'Failed to distribute parameters to: {reject_list}')
         results.update({_target: True for _target in accept_list})
 
-        _min_clients = (self.min_clients) - 1 if self.involve_aggregator else self.min_clients
-        if sum(results.values()) < _min_clients:
+        if sum(results.values()) < len(self.gethered):
             self.push_log('Task failed because of too few calculators getting ready')
             raise AggregationError(f'Too few calculators getting ready: {results}.')
         self.push_log(f'Distributed parameters to {sum(results.values())} calculators.')
@@ -755,7 +554,7 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         self.contractor.notify_ready_for_aggregation(round=self.current_round)
         self.push_log('Now waiting for executing calculation ...')
         accum_result, result_count = self._wait_for_calculation()
-        if result_count < self.min_clients:
+        if result_count < len(self.gethered) + int(self.involve_aggregator):
             self.push_log('Task failed because of too few calculation results gathered.')
             raise AggregationError(f'Too few results gathered: {result_count} copies.')
         self.push_log(f'Received {result_count} copies of calculation results.')
@@ -786,13 +585,13 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
             or self.current_round == self.max_rounds
         ):
             self.push_log('Begin to make a model test.')
-            self.test()
+            self.run_test()
             self.push_log('Finished a round of test.')
 
     def _wait_for_calculation(self) -> Tuple[Dict[str, torch.Tensor], int]:
         """Wait for every calculator finish its task or timeout."""
         result_count = 0
-        if self.id in self._calculators:
+        if self.involve_aggregator:
             self._execute_training()
             result_count += 1
             self.push_log(f'The aggregator ID: {self.id} obtained its calculation results.')
@@ -816,33 +615,20 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
                     accum_result[_key].add_(_new_state_dict[_key])
                 result_count += 1
                 self.push_log(f'Received calculation results from ID: {_event.source}')
-                if result_count >= len(self._calculators):
+                logger.info('======================================')
+                logger.info(f'{result_count=}')
+                logger.info(f'{len(self.gethered)=}')
+                logger.info(f'{int(self.involve_aggregator)=}')
+                logger.info(f'{(result_count == len(self.gethered) + int(self.involve_aggregator))=}')
+                if result_count == len(self.gethered) + int(self.involve_aggregator):
                     return accum_result, result_count
 
-            elif isinstance(_event, CheckinEvent):
-                self._handle_check_in(_event)
-
         return accum_result, result_count
 
     def _is_reach_max_rounds(self) -> bool:
         """Is the max rounds configuration reached."""
         return self.current_round >= self.max_rounds
 
-    def _select_calculators(self):
-        """Select calculators of a round."""
-        if len(self._participants) < self.min_clients:
-            raise AggregationError(f'too few participants: {len(self._participants)}')
-
-        if len(self._calculators) < self.min_clients:
-            candidates = [_peer for _peer in self._participants
-                          if _peer not in self._calculators]
-            random.shuffle(candidates)
-            self._calculators.extend(candidates[:self.min_clients - len(self._calculators)])
-
-        elif len(self._calculators) > self.max_clients:
-            random.shuffle(self._calculators)
-            self._calculators = self._calculators[:self.max_clients]
-
     def _save_model(self):
         """Save latest model state."""
         with open(self._ckpt_file, 'wb') as f:
@@ -867,16 +653,8 @@ class FedAvgScheduler(Scheduler, metaclass=ABCMeta):
         """
         self.push_log('Uploading task achievement and closing task ...')
 
-        # TODO remove metrics_files later
-        metrics_files = []
-        for _name, _metrics in self._metrics_bucket.items():
-            _file = f'{os.path.join(self._result_dir, _name)}.csv'
-            _metrics.to_csv(_file)
-            metrics_files.append(_file)
         report_file = os.path.join(self._result_dir, "report.zip")
         with ZipFile(report_file, 'w') as report_zip:
-            for _file in metrics_files:
-                report_zip.write(_file, os.path.basename(_file))
             for path, _, filenames in os.walk(self._log_dir):
                 rel_dir = os.path.relpath(path=path, start=self._result_dir)
                 rel_dir = rel_dir.lstrip('.')  # ./file => file
@@ -990,51 +768,32 @@ class FedSGDScheduler(FedAvgScheduler):
     """Implementation of FedSGD."""
 
     def __init__(self,
-                 min_clients: int,
-                 name: str = None,
                  max_rounds: int = 0,
                  calculation_timeout: int = 300,
-                 perf_bench_timeout: int = 30,
                  schedule_timeout: int = 30,
-                 log_rounds: int = 0,
-                 is_centralized: bool = True):
+                 log_rounds: int = 0):
         """Init.
 
         By now, IterableDataset with no length is not supported.
 
         Args:
-            min_clients:
-                Minimal number of calculators for each round.
-            name:
-                Default to the task ID.
             max_rounds:
                 Maximal number of training rounds.
             calculation_timeout:
                 Seconds to timeout for calculation in a round. Takeing off timeout
                 by setting its value to 0.
-            perf_bench_timeout:
-                Seconds to timeout for performing performance benchmark. Takeing off timeout
-                by setting its value to 0.
             schedule_timeout:
                 Seconds to timeout for process scheduling. Takeing off timeout
                 by setting its value to 0.
             log_rounds:
                 The number of rounds to run testing and log the result. Skip it
                 by setting its value to 0.
-            is_centralized:
-                If specify centralized, the aggregator will always be the initiator of the
-                task, otherwize a new aggregator is elected for each round during training.
         """
-        super().__init__(min_clients=min_clients,
-                         max_clients=sys.maxsize,
-                         name=name,
-                         max_rounds=max_rounds,
+        super().__init__(max_rounds=max_rounds,
                          merge_epochs=1,
                          calculation_timeout=calculation_timeout,
-                         perf_bench_timeout=perf_bench_timeout,
                          schedule_timeout=schedule_timeout,
-                         log_rounds=log_rounds,
-                         is_centralized=is_centralized)
+                         log_rounds=log_rounds)
 
     def _setup_context(self, id: str, task_id: str, is_initiator: bool = False):
         super()._setup_context(id=id, task_id=task_id, is_initiator=is_initiator)
@@ -1046,12 +805,7 @@ class FedSGDScheduler(FedAvgScheduler):
         try:
             num_samples = len(self.train_loader.dataset)
             self.push_log(f'train examples = {num_samples}')
-            if self.train_loader.batch_size != len(num_samples):
+            if self.train_loader.batch_size != num_samples:
                 raise ConfigError('batch size must be the total number of samples.')
         except TypeError:
             raise ConfigError('Does not support iterable dataset with no length.')
-
-    def _select_calculators(self):
-        """Select calculators of a round."""
-        if len(self._calculators) < self.min_clients:
-            raise AggregationError(f'too few calculators: {len(self._calculators)}')
