@@ -8,6 +8,7 @@ import base64
 import io
 import json
 import secrets
+from copy import deepcopy
 from dataclasses import dataclass
 from typing import Dict, List, Tuple
 
@@ -20,9 +21,9 @@ from ..scheduler import ConfigError, TaskComplete
 from ..secure import aes
 from ..secure.ecdhe import new_key_pair, new_shared_key
 from ..secure.shamir import SecretShare
-from .contractor import (CheckinEvent, FinishTaskEvent,
+from .contractor import (CheckinEvent, FinishTaskEvent, ResetRoundEvent,
                          UploadTrainingResultsEvent)
-from .fed_avg import AggregationError, FedAvgScheduler
+from .fed_avg import AggregationError, FedAvgScheduler, ResetRound
 from .secure_contractor import (AdvertiseKeysEvent, AdvertiseKeysResponseEvent,
                                 DistributeUser1ListEvent,
                                 DistributeUser2ListEvent,
@@ -193,17 +194,12 @@ class SecureFedAvgScheduler(FedAvgScheduler):
     _UNMASKING = 'unmasking'
 
     def __init__(self,
-                 min_clients: int,
-                 max_clients: int,
                  t: int,
-                 name: str = None,
                  max_rounds: int = 0,
                  merge_epochs: int = 1,
                  calculation_timeout: int = 300,
-                 perf_bench_timeout: int = 30,
                  schedule_timeout: int = 30,
-                 log_rounds: int = 0,
-                 is_centralized: bool = True):
+                 log_rounds: int = 0):
         """Init.
 
         Reference: https://eprint.iacr.org/2017/281.pdf
@@ -211,45 +207,28 @@ class SecureFedAvgScheduler(FedAvgScheduler):
         Because of conflicting with this Privacy-Preserving protocol, `involve aggregator`
         mode is not supported.
 
-        :args
-            :participants
-                ID list of all participants registered on the task.
-            :min_clients
-                Minimal number of calculators for each round, at least 3.
-            :max_clients
-                Maximal number of calculators for each round.
-            :t
+        Args:
+            t:
                 The threshold value t for restoring a shared secret.
-            :max_rounds
+            max_rounds:
                 Maximal number of training rounds.
-            :merge_epochs
+            merge_epochs:
                 The number of epochs to run before aggregation is performed.
-            :calculation_timeout
+            calculation_timeout:
                 Seconds to timeout for calculation in a round. Takeing off timeout
                 by setting its value to 0.
-            :perf_bench_timeout
-                Seconds to timeout for performing performance benchmark. Takeing off timeout
-                by setting its value to 0.
-            :schedule_timeout
+            schedule_timeout:
                 Seconds to timeout for process scheduling. Takeing off timeout
                 by setting its value to 0.
-            :log_rounds
+            log_rounds:
                 The number of rounds to run testing and log the result. Skip it
                 by setting its value to 0.
-            :is_centralized
-                If specify centralized, the aggregator will always be the initiator of the
-                task, otherwize a new aggregator is elected for each round during training.
         """
-        super().__init__(min_clients=min_clients,
-                         max_clients=max_clients,
-                         name=name,
-                         max_rounds=max_rounds,
+        super().__init__(max_rounds=max_rounds,
                          merge_epochs=merge_epochs,
                          calculation_timeout=calculation_timeout,
-                         perf_bench_timeout=perf_bench_timeout,
                          schedule_timeout=schedule_timeout,
-                         log_rounds=log_rounds,
-                         is_centralized=is_centralized)
+                         log_rounds=log_rounds)
         self._users_1: List[_User1Info] = []
         self._users_2: List[_User2Info] = []
         self._users_3: List[str] = []
@@ -265,19 +244,16 @@ class SecureFedAvgScheduler(FedAvgScheduler):
 
         self._t = t
 
-        self._validate_secure_config()
-
-    def _validate_secure_config(self):
-        # if self.min_clients < 4:
-        if self.min_clients < 3:  # TODO it should be 4
-            raise ConfigError('secure mode works only with at least 4 participants')
-        if self._t > self.min_clients - 1:
-            raise ConfigError('threshold t must be equal or less than min_clients - 1')
-
     def _setup_context(self, id: str, task_id: str, is_initiator: bool = False):
         super()._setup_context(id, task_id, is_initiator)
         self.contractor = SecureFedAvgContractor(task_id=self.task_id)
 
+        # if self._participants < 4:
+        if self.participants < 3:  # TODO it should be 4
+            raise ConfigError('secure mode works only with at least 4 participants')
+        if self._t > self.participants - 1:
+            raise ConfigError('threshold t must be less than the number of participants.')
+
     def _get_user_2_num(self):
         """Return the number of user_2.
 
@@ -291,11 +267,11 @@ class SecureFedAvgScheduler(FedAvgScheduler):
     def _process_aggregation(self):
         self._switch_status(self._ADVERTISE_KEYS)
         self.push_log('Initiate advertising keys.')
-        self.contractor.advertise_keys(round=self._round, calculators=self._calculators)
+        self.contractor.advertise_keys(round=self.current_round, calculators=self.gethered)
         self._wait_for_responses_of_advertising_keys()
         self.push_log(f'Gathered user_1 list: {[_user.peer_id for _user in self._users_1]}')
         list_data = [_user.encode() for _user in self._users_1]
-        self.contractor.distribute_user_1_list(round=self._round,
+        self.contractor.distribute_user_1_list(round=self.current_round,
                                                list_data=list_data,
                                                targets=[_user.peer_id for _user in self._users_1])
         self.push_log('Distributed user_1 list.')
@@ -305,21 +281,21 @@ class SecureFedAvgScheduler(FedAvgScheduler):
         self.push_log(f'Gathered user_2 list: {[_user.source_id for _user in self._users_2]}')
         list_data = [_user.encode() for _user in self._users_2]
         self.contractor.distribute_user_2_list(
-            round=self._round,
+            round=self.current_round,
             list_data=list_data,
             targets=list(set(_user.source_id for _user in self._users_2))
         )
         self.push_log('Distributed user_2 list.')
 
         self._switch_status(self._MASKED_INPUT_COLLECTION)
-        self.contractor.notify_ready_for_aggregation(round=self._round)
+        self.contractor.notify_ready_for_aggregation(round=self.current_round)
         self.push_log('Now waiting for executing calculation ...')
         accum_result, result_count = self._wait_for_calculation()
         self.push_log(f'Gathered user_3 list: {self._users_3}')
 
         self._switch_status(self._UNMASKING)
         self.push_log('Trying to unmask parameters ...')
-        self.contractor.start_unmasking(round=self._round,
+        self.contractor.start_unmasking(round=self.current_round,
                                         list_data=self._users_3,
                                         targets=self._users_3)
         self._wait_for_collecting_shares()
@@ -334,7 +310,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
         advertised_ids = []
         for _event in self.contractor.contract_events(timeout=self.schedule_timeout):
             if isinstance(_event, AdvertiseKeysResponseEvent):
-                if _event.round != self._round:
+                if _event.round != self.current_round:
                     continue
                 if _event.peer_id in advertised_ids:
                     continue
@@ -343,7 +319,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                                                 s_pk=_event.s_pk))
                 advertised_ids.append(_event.peer_id)
                 self.push_log(f'Received advertised keys from ID: {_event.peer_id}.')
-                if len(self._users_1) == len(self._calculators):
+                if len(self._users_1) == len(self.gethered):
                     break
 
             elif isinstance(_event, CheckinEvent):
@@ -360,7 +336,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
         shared_ids = []
         for _event in self.contractor.contract_events(timeout=self.schedule_timeout):
             if isinstance(_event, ShareKeysEvent):
-                if _event.round != self._round:
+                if _event.round != self.current_round:
                     continue
                 if _event.peer_id in shared_ids:
                     continue
@@ -384,16 +360,17 @@ class SecureFedAvgScheduler(FedAvgScheduler):
         accum_result = self.state_dict()
         for _key, _param in accum_result.items():
             if isinstance(_param, torch.Tensor):
-                _param.zero_()
+                accum_result[_key] = _param.cpu().zero_()
             else:
                 logger.warn(f'not Tensor value: {_key=}; {_param=}; {type(_param)=}')
 
         self.push_log('Waiting for training results ...')
         for _event in self.contractor.contract_events(timeout=self.calculation_timeout):
             if isinstance(_event, UploadTrainingResultsEvent):
-                if _event.target != self.id:
+                if self.id not in _event.target:
                     continue
-                training_result = self.data_channel.receive_stream(_event)
+                training_result = self.data_channel.receive_stream(apply_event=_event,
+                                                                   receiver=self.id)
                 buffer = io.BytesIO(training_result)
                 _new_state_dict = torch.load(buffer)
                 for _key in accum_result.keys():
@@ -421,7 +398,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
         self.push_log('Waiting for collecting secret shares ...')
         for _event in self.contractor.contract_events(timeout=self.schedule_timeout):
             if isinstance(_event, UploadSeedSharesEvent):
-                if _event.round != self._round:
+                if _event.round != self.current_round:
                     continue
 
                 share_info = [_SeedShareInfo.decode(_item) for _item in _event.list_data]
@@ -441,7 +418,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                     break
 
             if isinstance(_event, UploadSKSharesEvent):
-                if _event.round != self._round:
+                if _event.round != self.current_round:
                     continue
 
                 share_info = [_SKShareInfo.decode(_item) for _item in _event.list_data]
@@ -475,7 +452,13 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                                                sk_map=sk_map)
 
         for _key in accum_result.keys():
-            accum_result[_key].div_(result_count)
+            if accum_result[_key].dtype in (
+                torch.uint8, torch.int8, torch.int16, torch.int32, torch.int64
+            ):
+                logger.warn(f'average a int value may lose precision: {_key=}')
+                accum_result[_key].div_(result_count, rounding_mode='trunc')
+            else:
+                accum_result[_key].div_(result_count)
         self.load_state_dict(accum_result)
         self.push_log('Obtained a new version of parameters.')
 
@@ -580,7 +563,6 @@ class SecureFedAvgScheduler(FedAvgScheduler):
 
         self._switch_status(self._UPDATING)
         self._wait_for_updating_model()
-        self._save_model()
 
         self._switch_status(self._ADVERTISE_KEYS)
         self._wait_for_advertising_keys()
@@ -588,7 +570,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
 
         self._switch_status(self._SHARE_KEYS)
         shared_data = self._generate_shared_data()
-        self.contractor.share_keys(round=self._round,
+        self.contractor.share_keys(round=self.current_round,
                                    peer_id=self.id,
                                    list_data=shared_data,
                                    aggregator=self._aggregator)
@@ -598,23 +580,19 @@ class SecureFedAvgScheduler(FedAvgScheduler):
         self._switch_status(self._CALCULATING)
         self.push_log('Begin to run calculation ...')
         for _ in range(self.merge_epochs):
-            self.train()
-        state_dict = self.state_dict()
-        self.push_log('Local calculation complete.')
+            self.train_an_epoch()
 
+        self.push_log('Local calculation complete.')
         self._switch_status(self._MASKED_INPUT_COLLECTION)
-        masked_result = self._mask_state_dict(state_dict=state_dict)
+        masked_result = self._mask_state_dict()
 
         self._wait_for_uploading_model()
         buffer = io.BytesIO()
         torch.save(masked_result, buffer)
-        connection_timeout, timeout = self.data_channel_timeout
         self.push_log('Pushing local update to the aggregator ...')
         self.data_channel.send_stream(source=self.id,
-                                      target=self._aggregator,
-                                      data_stream=buffer.getvalue(),
-                                      connection_timeout=connection_timeout,
-                                      timeout=timeout)
+                                      target=[self._aggregator],
+                                      data_stream=buffer.getvalue())
         self.push_log('Successfully pushed local update to the aggregator.')
 
         self._switch_status(self._UNMASKING)
@@ -622,18 +600,18 @@ class SecureFedAvgScheduler(FedAvgScheduler):
 
         self._switch_status(self._CLOSING_ROUND)
         self._wait_for_closing_round()
-        self.push_log(f'ID: {self.id} finished training task of round {self._round}.')
+        self.push_log(f'ID: {self.id} finished training task of round {self.current_round}.')
 
     def _wait_for_advertising_keys(self):
         """Wait for the start signal and then advertise keys as a calculator."""
         self.push_log('Waiting for the signal to advertise keys ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, AdvertiseKeysEvent):
-                if _event.round != self._round or self.id not in _event.calculators:
+                if _event.round != self.current_round or self.id not in _event.calculators:
                     continue
                 self._c_sk, self._c_pk = new_key_pair()
                 self._s_sk, self._s_pk = new_key_pair()
-                self.contractor.respond_advertise_keys(round=self._round,
+                self.contractor.respond_advertise_keys(round=self.current_round,
                                                        aggregator=self._aggregator,
                                                        peer_id=self.id,
                                                        c_pk=self._c_pk,
@@ -642,6 +620,8 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                 return
             elif isinstance(_event, FinishTaskEvent):
                 raise TaskComplete()
+            elif isinstance(_event, ResetRoundEvent):
+                raise ResetRound()
 
     def _wait_for_user_1_list(self):
         """Wait for receiving the list of user_1."""
@@ -649,7 +629,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
         self.push_log('Waiting for receiving users_1 list ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, DistributeUser1ListEvent):
-                if _event.round != self._round:
+                if _event.round != self.current_round:
                     continue
                 self._users_1 = [_User1Info.decode(_user)
                                  for _user in _event.list_data]
@@ -670,6 +650,8 @@ class SecureFedAvgScheduler(FedAvgScheduler):
 
             elif isinstance(_event, FinishTaskEvent):
                 raise TaskComplete()
+            elif isinstance(_event, ResetRoundEvent):
+                raise ResetRound()
 
     def _generate_shared_data(self) -> List[str]:
         """Generate information for sharing depending on user_1 list."""
@@ -698,7 +680,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
         self.push_log('Waiting for receiving users_2 list ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, DistributeUser2ListEvent):
-                if _event.round != self._round:
+                if _event.round != self.current_round:
                     continue
                 self._users_2 = [_User2Info.decode(_user)
                                  for _user in _event.list_data]
@@ -720,8 +702,11 @@ class SecureFedAvgScheduler(FedAvgScheduler):
 
             elif isinstance(_event, FinishTaskEvent):
                 raise TaskComplete()
+            elif isinstance(_event, ResetRoundEvent):
+                raise ResetRound()
 
-    def _mask_state_dict(self, state_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
+    def _mask_state_dict(self) -> Dict[str, torch.Tensor]:
+        state_dict = deepcopy(self.state_dict())
         assert (
             state_dict and isinstance(state_dict, dict)
             and all(
@@ -733,6 +718,9 @@ class SecureFedAvgScheduler(FedAvgScheduler):
             )
         ), f'invalid state_dict data: {state_dict}'
 
+        # cuda tensor cannot operate with numpy tensor or cpu tensor
+        state_dict = {_key: _tensor.cpu() for _key, _tensor in state_dict.items()}
+
         self.push_log('Masking parameters ...')
         # in order to make every calculator's PRG generates masks in a same order
         sorted_keys = sorted(state_dict)
@@ -761,7 +749,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
         self.push_log('Waiting for unmasking signal ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, StartUnmaskingEvent):
-                if _event.round != self._round:
+                if _event.round != self.current_round:
                     continue
                 self._users_3 = _event.list_data
                 self._users_3.remove(self.id)
@@ -773,7 +761,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                                    for _user in self._users_2
                                    if _user.source_id in self._users_3]
                 self.contractor.upload_seed_shares(
-                    round=self._round,
+                    round=self.current_round,
                     list_data=[_info.encode() for _info in seed_share_list],
                     aggregator=self._aggregator
                 )
@@ -783,7 +771,7 @@ class SecureFedAvgScheduler(FedAvgScheduler):
                                  for _user in self._users_2
                                  if _user.source_id in lost_users]
                 self.contractor.upload_sk_shares(
-                    round=self._round,
+                    round=self.current_round,
                     list_data=[_info.encode() for _info in sk_share_list],
                     aggregator=self._aggregator
                 )
@@ -793,3 +781,5 @@ class SecureFedAvgScheduler(FedAvgScheduler):
 
             elif isinstance(_event, FinishTaskEvent):
                 raise TaskComplete()
+            elif isinstance(_event, ResetRoundEvent):
+                raise ResetRound()
